# -*- coding: utf-8 -*-


class Area:
    def __init__(self, model, study_area):
        """ """
        self.model = model
        self.geoms = study_area
        assert "xmin" in self.geoms
        assert "xmax" in self.geoms
        assert "ymin" in self.geoms
        assert "ymax" in self.geoms



# -*- coding: utf-8 -*-
"""This package is a framework for creating large-scale agent-based geographical models.

Submodules
==========


"""

__title__ = "honeybees"
__version__ = "1.0.6"
__email__ = "j.a.debruijn at outlook com"
__status__ = "development"



# -*- coding: utf-8 -*-
from typing import Tuple
from rasterio.transform import Affine
from rasterio.profiles import Profile
from numba import njit, prange
import numpy as np
from rasterio.windows import Window
from rasterio.crs import CRS
from rasterio.io import DatasetReader
from netCDF4 import Dataset, date2num
from datetime import datetime
import os
from typing import Union


@njit(cache=True)
def reproject_pixel(
    gt_from: tuple[float, float, float, float, float, float],
    gt_to: tuple[float, float, float, float, float, float],
    px_in: int,
    py_in: int,
) -> tuple[int, int]:
    """Reproject pixel.

    Args:
        gt_from: Orginal geotransformation.
        gt_to: Target geotransformation.
        px_in: Input xpixel.
        py_in: Input ypixel.
    """
    px_out = (gt_from[0] + gt_from[1] * px_in - gt_to[0]) / gt_to[1]
    py_out = (gt_from[3] + gt_from[5] * py_in - gt_to[3]) / gt_to[5]
    return int(px_out), int(py_out)


@njit(cache=True)
def pixel_to_coord(px: int, py: int, gt: tuple) -> Tuple[float, float]:
    """Converts pixel (x, y) to coordinate (lon, lat) for given geotransformation.
    Uses the upper left corner of the pixel. To use the center, add 0.5 to input pixel.

        Parameters:
            pixel: the pixel (x, y) that need to be transformed to coordinate
            gt: the geotransformation. Must be unrotated

        Returns:
            array: the coordinate (lon, lat)
    """
    if gt[2] + gt[4] == 0:
        lon = px * gt[1] + gt[0]
        lat = py * gt[5] + gt[3]
        return lon, lat
    else:
        raise ValueError("Cannot convert rotated maps")


@njit(cache=True, parallel=True)
def pixels_to_coords(pixels: np.ndarray, gt: tuple) -> np.ndarray:
    """Converts pixels (x, y) to coordinates (lon, lat) for given geotransformation.
    Uses the upper left corner of the pixels. To use the centers, add 0.5 to input pixels.

        Parameters:
            pixels: the pixels (x, y) that need to be transformed to coordinates (shape: 2, n)
            gt: the geotransformation. Must be unrotated

        Returns:
            array: the coordinates (lon, lat) (shape: 2, n)
    """
    assert pixels.shape[1] == 2
    if gt[2] + gt[4] == 0:
        coords = np.empty(pixels.shape, dtype=np.float64)
        for i in prange(coords.shape[0]):
            coords[i, 0] = pixels[i, 0] * gt[1] + gt[0]
            coords[i, 1] = pixels[i, 1] * gt[5] + gt[3]
        return coords
    else:
        raise ValueError("Cannot convert rotated maps")


@njit(cache=True)
def coord_to_pixel(
    coord: np.ndarray, gt: tuple[float, float, float, float, float, float]
) -> tuple[int, int]:
    """Converts coordinate to pixel (x, y) for given geotransformation.

    Parameters:
        coord: the coordinate (lon, lat) that need to be transformed to pixel
        gt: the geotransformation. Must be unrotated

    Returns:
        array: tuple of pixel (x, y)
    """
    if gt[2] + gt[4] == 0:
        px = (coord[0] - gt[0]) / gt[1]
        py = (coord[1] - gt[3]) / gt[5]
        return int(px), int(py)
    else:
        raise ValueError("Cannot convert rotated maps")


@njit(parallel=True)
def coords_to_pixels(
    coords, gt: tuple, dtype=np.uint32
) -> tuple[np.ndarray, np.ndarray]:
    """Converts array of coordinates to array of pixels for given geotransformation.

    Parameters:
        coords: the coordinates (lon, lat) that need to be transformed to pixels (shape: 2, n)
        gt: the geotransformation. Must be unrotated

    Returns:
        array: 2d-array of pixels per coordinate (shape: 2, n)
    """
    if gt[2] + gt[4] == 0:
        size = coords.shape[0]
        x_offset = gt[0]
        y_offset = gt[3]
        x_step = gt[1]
        y_step = gt[5]
        pxs = np.empty(size, dtype=dtype)
        pys = np.empty(size, dtype=dtype)
        for i in prange(size):
            pxs[i] = int((coords[i, 0] - x_offset) / x_step)
            pys[i] = int((coords[i, 1] - y_offset) / y_step)
        return pxs, pys
    else:
        raise ValueError("Cannot convert rotated maps")


@njit(parallel=False)
def sample_from_map(
    array: np.ndarray,
    coords: np.ndarray,
    gt: tuple[float, float, float, float, float, float],
) -> np.ndarray:
    """Sample coordinates from a map. Can handle multiple dimensions.

    Parameters:
        array: the map to sample from (2+n dimensions)
        coords: the coordinates used to sample (shape: 2, m)
        gt: the geotransformation. Must be unrotated

    Returns:
        array: values per coordinate
    """
    assert gt[2] + gt[4] == 0
    size = coords.shape[0]
    x_offset = gt[0]
    y_offset = gt[3]
    x_step = gt[1]
    y_step = gt[5]
    values = np.empty((size,) + array.shape[:-2], dtype=array.dtype)
    for i in prange(size):
        values[i] = array[
            ...,
            int((coords[i, 1] - y_offset) / y_step),
            int((coords[i, 0] - x_offset) / x_step),
        ]
    return values


@njit(
    parallel=False,
    cache=True,
)  # Writing to an array cannot be parallelized as race conditions would occur.
def write_to_array(
    array: np.ndarray,
    values: np.ndarray,
    coords: np.ndarray,
    gt: tuple[float, float, float, float, float, float],
):
    """Write values using coordinates to a map. If multiple coordinates map to a single cell,
    the values are added. The operation is inplace

        Parameters:
            array: the 2-dimensional array to write to
            values: the values to write (shape: n)
            coords: the coordinates of the values (shape: 2, n)
            gt: the geotransformation. Must be unrotated

        Returns:
            array: the array with the values added (operation is inplace)
    """
    assert values.size == coords.shape[0]
    assert gt[2] + gt[4] == 0
    size = values.size
    x_offset = gt[0]
    y_offset = gt[3]
    x_step = gt[1]
    y_step = gt[5]
    for i in range(size):
        array[
            int((coords[i, 1] - y_offset) / y_step),
            int((coords[i, 0] - x_offset) / x_step),
        ] += values[i]
    return array


def clip_to_xy_bounds(
    src: DatasetReader,
    profile: Profile,
    array: np.ndarray,
    xmin: int,
    xmax: int,
    ymin: int,
    ymax: int,
) -> np.ndarray:
    """Clips rasterio dataset to given bounds.

    Args:
        src: Rasterio dataset.
        profile: Rasterio profile of dataset.
        array: Array to clip.
        xmin: Minimum xbound.
        xmax: Maximum xbound.
        ymin: Minimum ybound.
        ymax: Maximum ybound.

    Returns:
        profile: Updated profile for cut array.
        array: Array data.
    """
    window = Window.from_slices(slice(ymin, ymax), slice(xmin, xmax))
    profile.update(
        {
            "transform": src.window_transform(window),
            "height": int(window.height),
            "width": int(window.width),
        }
    )
    return profile, array[ymin:ymax, xmin:xmax].copy()


def clip_to_other(
    array: np.ndarray, src_profile: Profile, other_profile: Profile
) -> tuple[np.ndarray, Profile]:
    """Clip array to rasterio profile.

    Args:
        array: Array to clip.
        src_profile: Rasterio profile of source array.
        other_profile: Rasterio profile of array to clip to.

    Returns:
        outarray: Clipped array.
        profile: Updated profile.
    """

    xmin = round(
        (other_profile["transform"].c - src_profile["transform"].c)
        / src_profile["transform"].a
    )
    assert abs(xmin) == xmin  # assert xmin is positive
    ymin = round(
        (other_profile["transform"].f - src_profile["transform"].f)
        / src_profile["transform"].e
    )
    assert abs(ymin) == ymin  # assert ymin is positive

    profile = dict(src_profile)
    profile.update(
        {
            "transform": other_profile["transform"],
            "height": other_profile["height"],
            "width": other_profile["width"],
        }
    )
    outarray = array[
        ymin : ymin + other_profile["height"], xmin : xmin + other_profile["width"]
    ].copy()
    return outarray, profile


def upscale(
    array: np.ndarray, src_profile: np.ndarray, factor: np.ndarray
) -> tuple[np.ndarray, Profile]:
    """ "Upscale array and rasterio profile by x amount.

    Args:
        array: Array to upscale.
        src_profile: Rasterio profile of source array.
        factor: Factor by which to upscale.

    Returns:
        array: Upscaled array.
        profile: Updated rasterio profile.
    """
    profile = dict(src_profile)
    transform = src_profile["transform"]
    profile["transform"] = Affine(
        transform.a / factor,
        transform.b,
        transform.c,
        transform.d,
        transform.e / factor,
        transform.f,
    )
    profile["height"] = src_profile["height"] * factor
    profile["width"] = src_profile["width"] * factor
    array = array.repeat(factor, axis=-2).repeat(factor, axis=-1)
    return array, profile


class NetCDFHasEmtpyLayersException(Exception):
    """Raised when one or more layers are empty"""

    pass


class CreateNetCDF:
    """Class to create easy create a spatial NetCDF-file.

    Args:
        fn: Filepath.
        title: Title of NetCDF.
        source: Source of data.
        institution: Institution where data was created.
        n_timesteps: Number of timesteps in NetCDF. None if NetCDF should not have time dimension.
        lons: Array of longitudes.
        lats: Array of latitudes.
        epsg: EPSG code.
        varname: Name of variable to store.
        units: Unit of variable.
        dtype: NetCDF dtype. See NetCDF documentation for valid dtypes: https://unidata.github.io/netcdf4-python/
        chunksizes: Sizes of datachunks. (time, lat, lon) if time dimension exists, otherwise (lat, lon).
        fill_value: Value to represent nodata.
        compression_level: NetCDF compression level (1-9)
        comment: Optional dataset comment.
    """

    def __init__(
        self,
        fp: str,
        title: str,
        source: str,
        institution: str,
        n_timesteps: Union[None, int],
        lons: np.ndarray,
        lats: np.ndarray,
        epsg: int,
        varname: str,
        units: str,
        dtype: str,
        chunksizes: tuple[int],
        fill_value: Union[float, int],
        compression_level: int,
        comment: str = None,
    ) -> None:
        self.fp = fp
        self.ds = Dataset(self.fp, "w", format="NETCDF4")
        self.ds.set_always_mask(False)

        self.ds.date_created = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        self.ds.source = source
        self.ds.institution = institution
        self.ds.title = title
        self.ds.Conventions = "CF-1.6"
        if comment:
            self.ds.comment = comment

        if n_timesteps:
            self.ds.createDimension("time", n_timesteps)
        self.ds.createDimension("lat", len(lats))
        self.ds.createDimension("lon", len(lons))

        lat = self.ds.createVariable("lat", "f8", ("lat",))
        lat.standard_name = "latitude"
        lat.units = "degrees_north"
        lat.axis = "Y"
        lat[:] = lats

        lon = self.ds.createVariable("lon", "f8", ("lon",))
        lon.standard_name = "longitude"
        lon.units = "degrees_east"
        lon.axis = "X"
        lon[:] = lons

        crs = self.ds.createVariable("spatial_ref", "i4")
        crs.spatial_ref = CRS.from_epsg(epsg).to_wkt()

        if n_timesteps:
            self.times = self.ds.createVariable("time", "f4", ("time",))
            self.times.standard_name = "time"
            self.times.long_name = "time"
            self.times.units = "hours since 1970-01-01 00:00:00"
            self.times.calendar = "gregorian"

        self.values = self.ds.createVariable(
            varname,
            datatype=dtype,
            dimensions=("time", "lat", "lon") if n_timesteps else ("lat", "lon"),
            chunksizes=chunksizes,
            zlib=bool(compression_level),
            complevel=compression_level,  # ignored if zlib is False
            contiguous=False,  # do not neccesarily store contiguous on disk
            fill_value=fill_value,
        )
        self.values.standard_name = varname
        self.values.units = units
        self.current_timestep = 0

    def write(self, values: np.ndarray, dt: datetime = None) -> None:
        """Write layer to file.

        Args:
            values: Values to write.
            dt: Optional datetime to write. If NetCDF has no time dimension, do not write.
        """
        if dt:
            assert dt >= datetime(1970, 1, 1)
            self.times[self.current_timestep] = date2num(
                dt, units=self.times.units, calendar=self.times.calendar
            )
            self.values[self.current_timestep, :, :] = values
            self.current_timestep += 1
        else:
            assert not hasattr(self, "times")
            self.values[:, :] = values

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Check if NetCDF has no empty layers on exit."""
        if hasattr(self, "times"):
            checklayer = self.times
        else:
            checklayer = self.values

        if np.ma.is_masked(checklayer[:]):
            self.ds.close()
            os.remove(self.fp)
            raise NetCDFHasEmtpyLayersException
        else:
            self.ds.close()
            if exc_type is not None:
                os.remove(self.fp)



# -*- coding: utf-8 -*-
"""
Chart Module
============

Module for drawing live-updating line charts using Charts.js

"""

import json
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from honeybees.visualization.ModularVisualization import VisualizationElement


class ChartModule(VisualizationElement):
    """Each chart can visualize one or more model-level series as lines
     with the data value on the Y axis and the step number as the X axis.

    At the moment, each call to the render method returns a list of the most
    recent values of each series.

    Attributes:
        series: A list of dictionaries containing information on series to
                plot. Each dictionary must contain (at least) the "name" and
                "color" keys. The "name" value must correspond to a
                model-level series collected by the model's DataCollector, and
                "color" must have a valid HTML color.
        canvas_height, canvas_width: The width and height to draw the chart on
                                     the page, in pixels. Default to 200 x 500
        data_collector_name: Name of the DataCollector object in the model to
                             retrieve data from.
        template: "chart_module.html" stores the HTML template for the module.


    Example:
        schelling_chart = ChartModule([{"name": "happy", "color": "Black"}],
                                      data_collector_name="datacollector")

    TODO:
        Have it be able to handle agent-level variables as well.

        More Pythonic customization; in particular, have both series-level and
        chart-level options settable in Python, and passed to the front-end
        the same way that "color" is currently.

    """

    package_includes = ["Chart.min.js", "ChartModule.js"]

    def __init__(self, series, canvas_height=200, canvas_width=500):
        """
        Create a new line chart visualization.

        Args:
            series: A list of dictionaries containing series names and
                    HTML colors to chart them in, e.g.
                    [{"name": "happy", "color": "Black"},]
            canvas_height, canvas_width: Size in pixels of the chart to draw.
            data_collector_name: Name of the DataCollector to use.
        """
        self.series = series
        self.canvas_height = canvas_height
        self.canvas_width = canvas_width

        # if first series has a color, assert all series have a color
        # if first series has no color, assure remaining series have no color and
        # assign colors from color wheel
        if self.series:
            if "color" in self.series[0]:
                assert all("color" in series for series in self.series)
            else:
                assert all("color" not in series for series in self.series)
                series_length = len(self.series)
                color_map = plt.get_cmap("gist_rainbow")
                for i, series in enumerate(self.series):
                    color = mcolors.rgb2hex(color_map(i / series_length))
                    series["color"] = color

        series_json = json.dumps(self.series)
        new_element = "new ChartModule({}, {},  {})"
        new_element = new_element.format(series_json, canvas_width, canvas_height)
        self.js_code = "elements.push(" + new_element + ");"

        self.reset()

    def render(self, model, update):
        if not update:
            self.current_chart_x_index = 0

        xs = model.reporter.timesteps[self.current_chart_x_index :]
        if model.timestep_length.days >= 1:
            dateformat = "%d %b %Y"
        else:
            dateformat = "%d %b %Y %H:%M"
        xs = [x.strftime(dateformat) for x in xs]

        ys = []
        for s in self.series:
            name = s["name"]
            yvar = model.reporter.variables[name]
            if "ID" in s:
                yvar = yvar[s["ID"]]
            yvar = yvar[self.current_chart_x_index :]
            ys.append(yvar)  # Latest values
            if len(xs) != len(yvar):
                if len(yvar) == 0:
                    model.logger.info(
                        f"Variable '{name}' cannot be shown in chart because it is not found in reporter."
                    )
                else:
                    model.logger.info(
                        f"Variable '{name}' cannot be shown in chart for some timesteps because it is not found in reporter for all requested timesteps."
                    )

        self.current_chart_x_index += len(xs)

        return {"xs": xs, "ys": ys}

    def reset(self):
        self.current_chart_x_index = 0



# -*- coding: utf-8 -*-
import numpy as np
import numba as nb
from typing import Union

from honeybees.library import geohash


@nb.njit(cache=True)
def find_neighbor_search_count_index(search_counts, search_count_index, r):
    for neighbor_search_counts_index in range(search_count_index, -1, -1):
        if r >= search_counts[neighbor_search_counts_index, 2]:
            neighbor_search_counts_index += 1
            break
    return neighbor_search_counts_index


@nb.njit(parallel=True, locals={"r": nb.uint32}, cache=True)
def find_neighbors_numba(
    search_target_unique_hashcodes,
    search_source_unique_hashcodes,
    search_target_hashcodes_sort_indices,
    cumulative_count_per_search_hashcode,
    search_source_cumulative_counts_per_hashcode,
    map_search_source_to_target,
    reverse_search_source_hashcodes_sort_indices,
    n_neighbor,
    radius,
    bits,
    dtype,
    minx,
    maxx,
    miny,
    maxy,
    grid,
):
    # TODO: Fix the issue with set type inference in numba. This is only important for
    # n_neighbor > 20. The current implementation is not efficient for n_neighbor > 20.
    nanvalue = np.iinfo(dtype).max  # use the maximum value of uint as nanvalue
    neighbors = np.full(
        (map_search_source_to_target.size, n_neighbor), nanvalue, dtype=dtype
    )
    n_unique_hashcodes = search_target_unique_hashcodes.size
    for u in nb.prange(search_source_unique_hashcodes.size):
        search_source_hashcode = search_source_unique_hashcodes[u]
        x, y = geohash.decode(
            search_source_hashcode, bits, minx=minx, maxx=maxx, miny=miny, maxy=maxy
        )
        shifts = geohash.get_shifts(x, y, radius, bits, minx, maxx, miny, maxy, grid)
        neighbor_geohashes = geohash.shift_multiple(
            search_source_hashcode, bits, shifts
        )
        neighbor_geohashes = np.sort(neighbor_geohashes)
        search_counts = np.zeros((neighbor_geohashes.size, 3), dtype=dtype)

        index = -1
        search_count_index = 0
        previous_neighbor_geohash = 0
        for neighbor_geohash in neighbor_geohashes:
            if neighbor_geohash == previous_neighbor_geohash + 1:
                if index == n_unique_hashcodes:
                    break
                if search_target_unique_hashcodes[index] == neighbor_geohash:
                    search_counts[search_count_index, 1] += (
                        cumulative_count_per_search_hashcode[index + 1]
                        - cumulative_count_per_search_hashcode[index]
                    )
                    index += 1
            else:
                if search_counts[search_count_index, 1] != 0:
                    search_count_index += 1
                index = np.searchsorted(
                    search_target_unique_hashcodes, neighbor_geohash
                )
                if index == n_unique_hashcodes:
                    break
                search_counts[search_count_index, 0] = (
                    cumulative_count_per_search_hashcode[index]
                )
                if search_target_unique_hashcodes[index] == neighbor_geohash:
                    search_counts[search_count_index, 1] += (
                        cumulative_count_per_search_hashcode[index + 1]
                        - cumulative_count_per_search_hashcode[index]
                    )
                    index += 1

            if neighbor_geohash == search_source_hashcode:
                self_index = search_count_index
            previous_neighbor_geohash = neighbor_geohash

        if search_counts[search_count_index, 1] == 0:
            search_count_index -= 1
        search_counts = search_counts[: search_count_index + 1, :]
        search_counts[:, 2] = np.cumsum(search_counts[:, 1])

        n_neighbors = search_counts[:, 1].sum() - 1  # exclude self (not a neighbor)
        agent_neighbors = np.full(n_neighbor, nanvalue, dtype=dtype)
        if n_neighbors > n_neighbor:
            # To understand the algorithm need to realize that you choose a random value
            # from a range that is smaller than the full range, and then after each value, you extend
            # that to include one more. In the event of a collision, you can safely insert the max because it was never
            # possible to include it before.

            # The chances of a collision increase at the same rate that the number of values decreases,
            # so the probability of any one number being in the result is not skewed, or biased.
            # from: https://codereview.stackexchange.com/questions/61338/generate-random-numbers-without-repetitions
            for neighbor_index_sorted in range(
                search_source_cumulative_counts_per_hashcode[u],
                search_source_cumulative_counts_per_hashcode[u + 1],
            ):
                neighbor_index = reverse_search_source_hashcodes_sort_indices[
                    neighbor_index_sorted
                ]
                search_target_agent_index = (
                    map_search_source_to_target[neighbor_index_sorted]
                    - search_counts[self_index, 0]
                    + search_counts[:self_index, 1].sum()
                )
                J = n_neighbors - n_neighbor

                # if n_neighbor > 20:
                #     selected_agents = set()

                for i in range(n_neighbor):
                    r = np.random.randint(0, J + 1)
                    if (
                        r >= search_target_agent_index
                    ):  # ensure we do not pick the agent itself and include the last agent
                        r += 1

                    neighbor_search_counts_index = find_neighbor_search_count_index(
                        search_counts, search_count_index, r
                    )
                    if neighbor_search_counts_index == 0:
                        neighbor = search_counts[neighbor_search_counts_index, 0] + r
                    else:
                        neighbor = (
                            search_counts[neighbor_search_counts_index, 0]
                            + r
                            - search_counts[neighbor_search_counts_index - 1, 2]
                        )
                    # if n_neighbor > 20:
                    #     if neighbor in selected_agents:
                    #         r = int(J)
                    #         if (
                    #             r >= search_target_agent_index
                    #         ):  # ensure we do not pick the agent itself and include the last agent
                    #             r += 1
                    #         neighbor_search_counts_index = (
                    #             find_neighbor_search_count_index(
                    #                 search_counts, search_count_index, r
                    #             )
                    #         )
                    #         if neighbor_search_counts_index == 0:
                    #             neighbor = (
                    #                 search_counts[neighbor_search_counts_index, 0] + r
                    #             )
                    #         else:
                    #             neighbor = (
                    #                 search_counts[neighbor_search_counts_index, 0]
                    #                 + r
                    #                 - search_counts[neighbor_search_counts_index - 1, 2]
                    #             )
                    #     selected_agents.add(neighbor)
                    # else:
                    for k in range(i):
                        if agent_neighbors[k] == neighbor:
                            r = int(J)
                            if (
                                r >= search_target_agent_index
                            ):  # ensure we do not pick the agent itself and include the last agent
                                r += 1
                            neighbor_search_counts_index = (
                                find_neighbor_search_count_index(
                                    search_counts, search_count_index, r
                                )
                            )
                            if neighbor_search_counts_index == 0:
                                neighbor = (
                                    search_counts[neighbor_search_counts_index, 0] + r
                                )
                            else:
                                neighbor = (
                                    search_counts[neighbor_search_counts_index, 0]
                                    + r
                                    - search_counts[neighbor_search_counts_index - 1, 2]
                                )
                            break
                    agent_neighbors[i] = neighbor

                    J += 1
                for f in range(agent_neighbors.size):
                    if agent_neighbors[f] != nanvalue:
                        neighbors[neighbor_index, f] = (
                            search_target_hashcodes_sort_indices[agent_neighbors[f]]
                        )
                agent_neighbors.fill(nanvalue)
        else:
            for neighbor_index_sorted in range(
                search_source_cumulative_counts_per_hashcode[u],
                search_source_cumulative_counts_per_hashcode[u + 1],
            ):
                neighbor_index = reverse_search_source_hashcodes_sort_indices[
                    neighbor_index_sorted
                ]
                search_target_agent_index = map_search_source_to_target[
                    neighbor_index_sorted
                ]
                n = 0
                for m in range(search_counts.shape[0]):
                    for o in range(search_counts[m, 1]):
                        neighbor = search_counts[m, 0] + o
                        if neighbor != search_target_agent_index:
                            agent_neighbors[n] = neighbor
                            n += 1
                for f in range(agent_neighbors.size):
                    if agent_neighbors[f] != nanvalue:
                        neighbors[neighbor_index, f] = (
                            search_target_hashcodes_sort_indices[agent_neighbors[f]]
                        )
                agent_neighbors.fill(nanvalue)
    return neighbors


def process_hashcodes(
    location_hashcodes: np.ndarray,
    search_ids: Union[np.ndarray, None],
    search_target_ids: Union[np.ndarray, None],
    dtype: type,
) -> tuple[
    np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray
]:
    """
    Args:
        location_hashcodes: Geohashes of all agents.
        search_ids: Ids of the agents to find neighbors for. If None, search all agents.
        dtype: Datatype to use to hold indices.

    Returns:
        search_target_unique_hashcodes: Array of unique geohashes of agents.
        search_source_unique_hashcodes: Array of unique geohashes of agents to find neighbors for.
        search_target_hashcodes_sort_indices: Indices that can be used to sort the `location_hashcodes`.
        cumulative_count_per_search_hashcode: Cumulative count of the unique search hashcodes.
        map_search_source_to_target:
        search_source_hashcodes_sort_indices: Indices that can be used to sort the `location_hashcodes`.
    """
    if search_ids is None:
        search_ids = np.arange(location_hashcodes.size, dtype=np.int64)
    if search_target_ids is None:
        search_target_ids = np.arange(location_hashcodes.size, dtype=np.int64)

    assert isinstance(search_ids, np.ndarray)
    assert isinstance(search_target_ids, np.ndarray)

    search_target_hashcodes = location_hashcodes[search_target_ids]
    search_target_hashcodes_sort_indices = np.argsort(search_target_hashcodes).astype(
        dtype
    )
    sorted_search_target_hashcodes = search_target_hashcodes[
        search_target_hashcodes_sort_indices
    ]
    search_target_unique_hashcodes, search_target_unique_index = np.unique(
        sorted_search_target_hashcodes, return_index=True
    )
    search_target_unique_index = search_target_unique_index.astype(dtype)
    cumulative_count_per_search_target_hashcode = np.append(
        search_target_unique_index, sorted_search_target_hashcodes.size
    )

    # search_target_hashcodes_sort_indices = np.argsort(location_hashcodes).astype(dtype)
    # sorted_search_hashcodes = location_hashcodes[search_target_hashcodes_sort_indices]
    # search_target_unique_hashcodes, search_unique_index = np.unique(sorted_search_hashcodes, return_index=True)
    # search_unique_index = search_unique_index.astype(dtype)
    # cumulative_count_per_search_hashcode = np.append(search_unique_index, sorted_search_hashcodes.size)

    search_source_hashcodes = location_hashcodes[search_ids]
    search_source_hashcodes_sort_indices = np.argsort(search_source_hashcodes).astype(
        dtype
    )
    search_source_ids_sorted_by_hashcode = search_ids[
        search_source_hashcodes_sort_indices
    ]
    search_source_sorted_hashcodes = search_source_hashcodes[
        search_source_hashcodes_sort_indices
    ]
    search_source_unique_hashcodes, search_source_unique_index = np.unique(
        search_source_sorted_hashcodes, return_index=True
    )
    search_source_unique_index = search_source_unique_index.astype(dtype)
    search_source_cumulative_counts_per_hashcode = np.append(
        search_source_unique_index, search_source_sorted_hashcodes.size
    )

    inverse_idx = np.empty_like(
        location_hashcodes
    )  # should be good, we are mapping targets to sources
    inverse_idx[np.argsort(location_hashcodes).astype(dtype)] = np.arange(
        0, inverse_idx.size
    )
    inverse_idx[~np.isin(inverse_idx, search_target_ids)] = -1
    map_search_source_to_target = inverse_idx[search_source_ids_sorted_by_hashcode]

    # inverse_idx = np.empty_like(search_target_hashcodes_sort_indices)
    # inverse_idx[search_target_hashcodes_sort_indices] = np.arange(0, inverse_idx.size)
    # map_search_source_to_target = inverse_idx[search_ids_sorted_by_hashcode]

    return (
        search_target_unique_hashcodes,
        search_source_unique_hashcodes,
        search_target_hashcodes_sort_indices,
        cumulative_count_per_search_target_hashcode,
        search_source_cumulative_counts_per_hashcode,
        map_search_source_to_target,
        search_source_hashcodes_sort_indices,
    )


def find_neighbors(
    locations: np.ndarray,
    radius: Union[float, int],
    n_neighbor: int,
    bits: int,
    minx: Union[float, int] = -180,
    maxx: Union[float, int] = 180,
    miny: Union[float, int] = -90,
    maxy: Union[float, int] = 90,
    grid: str = "longlat",
    search_ids: Union[None, np.ndarray] = None,
    search_target_ids: Union[np.ndarray, None] = None,
) -> np.ndarray:
    """Finds neighbouring agents for given agents.

    Args:
        locations: Either a 2-dimensional array of x and y locations for agents, or a 1-dimensional array of geohashes.
        radius: Search radius. Specified in meters for grid=longlat, specified in map units for orthogonal grids.
        n_neighbors: Maxiumum number of neighbors to find for each agent.
        bits: Number of bits to use for geohashes.
        minx: Minimum x-value of the entire relevant space.
        maxx: Maximum x-value of the entire relevant space.
        miny: Minimum y-value of the entire relevant space.
        miny: Maximum y-value of the entire relevant space.
        grid: Type of grid. Can be 'longlat' or 'orthogonal'.
        search_ids: Only search for neighbors of agent with search_ids.
        search_target_ids: Only search neighbors with search_target_ids.

    Returns:
        neighbors: 2-dimensional NumPy array. The first dimension represents the results for each of the agents, the second dimension the indices of the neighbors for each of those agents. -1 represents no neighbor (possible only if less than the requested number of neighbors are available in the search radius.)
    """
    assert grid in ("longlat", "orthogonal")
    if locations.ndim == 2:
        assert locations.shape[1] == 2
        location_hashcodes = geohash.encode_locations(locations, minx, maxx, miny, maxy)
        location_hashcodes = geohash.reduce_precision(
            location_hashcodes, bits, inplace=True
        )
    elif locations.ndim == 1:
        location_hashcodes = geohash.reduce_precision(locations, bits, inplace=False)
        assert location_hashcodes.dtype == np.int64
    else:
        raise ValueError(
            "locations must be either 2d-array with lon,lat values or 1d-array of geohashes"
        )

    # test if number of locations is not larger than the maximum value an uint32 can hold. -1 because the max value is used as nan
    if location_hashcodes.size <= np.iinfo(np.uint32).max - 1:
        dtype = np.uint32
    else:
        dtype = np.uint64

    (
        search_target_unique_hashcodes,
        search_source_unique_hashcodes,
        search_target_hashcodes_sort_indices,
        cumulative_count_per_search_hashcode,
        search_source_cumulative_counts_per_hashcode,
        map_search_source_to_target,
        search_source_hashcodes_sort_indices,
    ) = process_hashcodes(location_hashcodes, search_ids, search_target_ids, dtype)

    neighbours = find_neighbors_numba(
        search_target_unique_hashcodes=search_target_unique_hashcodes,
        search_source_unique_hashcodes=search_source_unique_hashcodes,
        search_target_hashcodes_sort_indices=search_target_hashcodes_sort_indices,
        cumulative_count_per_search_hashcode=cumulative_count_per_search_hashcode,
        search_source_cumulative_counts_per_hashcode=search_source_cumulative_counts_per_hashcode,
        map_search_source_to_target=map_search_source_to_target,
        reverse_search_source_hashcodes_sort_indices=search_source_hashcodes_sort_indices,
        n_neighbor=n_neighbor,
        radius=radius,
        bits=bits,
        dtype=dtype,
        minx=minx,
        maxx=maxx,
        miny=miny,
        maxy=maxy,
        grid=grid,
    )

    if search_target_ids is not None:
        nanvalue = np.iinfo(dtype).max
        reindex_neighbours = np.take(
            search_target_ids, neighbours, out=np.empty_like(neighbours), mode="clip"
        )
        reindex_neighbours[neighbours == nanvalue] = nanvalue
        neighbours = reindex_neighbours

    return neighbours


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches

    n_locations = 1000
    radius = 5000
    n_neighbor = 5
    bits = 29

    locations = np.c_[
        np.random.uniform(33, 34, n_locations), np.random.uniform(-6, -5, n_locations)
    ]
    search_ids = np.random.choice(np.arange(0, n_locations), 50, replace=False)

    counts = np.zeros(n_locations, dtype=np.int32)
    for i in range(1000):
        neighors = find_neighbors(
            locations, radius, n_neighbor, bits, search_ids=search_ids
        )
        for neighbor in neighors.reshape(neighors.size):
            if neighbor != 4294967295:
                counts[neighbor] += 1

    window_width, window_height = geohash.window(bits)

    _, ax = plt.subplots()

    for agent, agent_neighbors in zip(search_ids, neighors):
        agent_neighbors = agent_neighbors[agent_neighbors != 4294967295]
        assert np.unique(agent_neighbors).size == agent_neighbors.size
        geohash_coord = geohash.decode(
            geohash.encode_precision(*locations[agent], bits), bits
        )
        shifts = geohash.get_shifts(geohash_coord[0], geohash_coord[1], radius, bits)
        neighbor_geohashes = geohash.shift_multiple(
            geohash.encode_precision(*locations[agent], bits), bits, shifts
        )
        neighbor_geohashes = np.sort(neighbor_geohashes)
        for j, neighbor_geohash in enumerate(neighbor_geohashes):
            patch = mpatches.Rectangle(
                geohash.decode(neighbor_geohash, bits),
                window_width,
                window_height,
                facecolor="orange",
                edgecolor="black",
                alpha=j / neighbor_geohashes.size * 0.5 + 0.1,
            )
            ax.add_patch(patch)

        loc = locations[agent]
        for neighbor in agent_neighbors:
            assert agent != neighbor
            neighborloc = locations[neighbor]
            ax.plot([loc[0], neighborloc[0]], [loc[1], neighborloc[1]])

    for x, y, count in zip(locations[:, 0], locations[:, 1], counts):
        ax.annotate(str(count), (x, y))

    ax.scatter(locations[:, 0], locations[:, 1], s=4)
    plt.show()



# -*- coding: utf-8 -*-
"""
ModularServer
=============

A visualization server which renders a model via one or more elements.

The concept for the modular visualization server as follows:
A visualization is composed of VisualizationElements, each of which defines how
to generate some visualization from a model instance and render it on the
client. VisualizationElements may be anything from a simple text display to
a multilayered HTML5 canvas.

The actual server is launched with one or more VisualizationElements;
it runs the model object through each of them, generating data to be sent to
the client. The client page is also generated based on the JavaScript code
provided by each element.

This file consists of the following classes:

VisualizationElement: Parent class for all other visualization elements, with
                      the minimal necessary options.
PageHandler: The handler for the visualization page, generated from a template
             and built from the various visualization elements.
SocketHandler: Handles the websocket connection between the client page and
                the server.
ModularServer: The overall visualization application class which stores and
               controls the model and visualization instance.


ModularServer should *not* need to be subclassed on a model-by-model basis; it
should be primarily a pass-through for VisualizationElement subclasses, which
define the actual visualization specifics.

For example, suppose we have created two visualization elements for our model,
called canvasvis and graphvis; we would launch a server with:

    server = ModularServer(MyModel, [canvasvis, graphvis], name="My Model")
    server.launch()

The client keeps track of what step it is showing. Clicking the Step button in
the browser sends a message requesting the viz_state corresponding to the next
step position, which is then sent back to the client via the websocket.

The websocket protocol is as follows:
Each message is a JSON object, with a "type" property which defines the rest of
the structure.

Server -> Client:
    Send over the model state to visualize.
    Model state is a list, with each element corresponding to a div; each div
    is expected to have a render function associated with it, which knows how
    to render that particular data. The example below includes two elements:
    the first is data for a CanvasGrid, the second for a raw text display.

    {
    "type": "viz_state",
    "data": [{0:[ {"Shape": "circle", "x": 0, "y": 0, "r": 0.5,
                "color": "#AAAAAA", "Filled": "true", "Layer": 0,
                "text": 'A', "text_color": "white" }]},
            "Shape Count: 1"]
    }

    Informs the client that the model is over.
    {"type": "end"}

    Informs the client of the current model's parameters
    {
    "type": "model_params",
    "params": 'dict' of model params, (i.e. {arg_1: val_1, ...})
    }

Client -> Server:
    Reset the model.
    TODO: Allow this to come with parameters
    {
    "type": "reset"
    }

    Get a given state.
    {
    "type": "get_step",
    "step:" index of the step to get.
    }

    Submit model parameter updates
    {
    "type": "submit_params",
    "param": name of model parameter
    "value": new value for 'param'
    }

    Get the model's parameters
    {
    "type": "get_params"
    }

"""

import os
import tornado.autoreload
import tornado.ioloop
import tornado.web
import tornado.websocket
import tornado.escape
import tornado.gen
import webbrowser

from honeybees.visualization.UserParam import UserSettableParameter


class VisualizationElement:
    """
    Defines an element of the visualization.

    Attributes:
        package_includes: A list of external JavaScript files to include that
                          are part of the honeybees packages.
        local_includes: A list of JavaScript files that are local to the
                        directory that the server is being run in.
        js_code: A JavaScript code string to instantiate the element.

    Methods:
        render: Takes a model object, and produces JSON data which can be sent
                to the client.

    """

    package_includes = []
    local_includes = []
    js_code = ""
    render_args = {}

    def __init__(self):
        pass

    def render(self, model, update):
        """Build visualization data from a model object.

        Args:
            model: A model object

        Returns:
            A JSON-ready object.

        """
        return "<b>VisualizationElement goes here</b>."

    def reset(self):
        pass


# =============================================================================
# Actual Tornado code starts here:


class PageHandler(tornado.web.RequestHandler):
    """Handler for the HTML template which holds the visualization."""

    def get(self):
        self.render(
            "modular_template.html",
            port=self.application.port,
            model_name=self.application.model_name,
            description=self.application.description,
            timesteps=self.application.timesteps,
            package_includes=self.application.package_includes,
            local_includes=self.application.local_includes,
            scripts=self.application.js_code,
        )


class SocketHandler(tornado.websocket.WebSocketHandler):
    """Handler for websocket."""

    def open(self):
        if self.application.verbose:
            print("Socket opened!")
        if not hasattr(self.application, "model"):
            self.write_message(
                {"type": "model_params", "params": self.application.user_params}
            )
        else:
            self.write_message({"type": "redraw"})

    def check_origin(self, origin):
        return True

    def viz_state_message(self, update=False):
        return {"type": "viz_state", "data": self.application.render_model(update)}

    @property
    def current_time(self):
        return {"type": "time", "data": self.application.model.current_time_fmt}

    def on_message(self, message):
        """Receiving a message from the websocket, parse, and act accordingly."""
        if self.application.verbose:
            print(message)
        msg = tornado.escape.json_decode(message)

        if msg["type"] == "step":
            if not self.application.model.running:
                self.write_message({"type": "end"})
            else:
                self.application.model.step(step_size=msg["n"])
                self.write_message(self.viz_state_message(update=True))
                self.write_message(self.current_time)

        elif msg["type"] == "reset":
            self.application.reset_model()
            self.write_message(self.viz_state_message(update=False))
            self.write_message(self.current_time)

        elif msg["type"] == "redraw":
            self.write_message(self.viz_state_message(update=False))

        elif msg["type"] == "change_map_variable":
            self.application.set_background_variable(variable_name=msg["variable"])
            self.write_message(self.viz_state_message(update=True))

        elif msg["type"] == "submit_params":
            param = msg["param"]
            value = msg["value"]

            # Is the param editable?
            if param in self.application.user_params:
                if isinstance(
                    self.application.model_kwargs[param], UserSettableParameter
                ):
                    self.application.model_kwargs[param].value = value
                else:
                    self.application.model_kwargs[param] = value

        else:
            if self.application.verbose:
                print("Unexpected message!")


class ModularServer(tornado.web.Application):
    """Main visualization application."""

    verbose = False

    port = 8521  # Default port to listen on
    max_steps = 100000

    # Handlers and other globals:
    page_handler = (r"/", PageHandler)
    socket_handler = (r"/ws", SocketHandler)
    static_handler = (
        r"/static/(.*)",
        tornado.web.StaticFileHandler,
        {"path": os.path.dirname(__file__) + "/templates"},
    )
    local_handler = (
        r"/local/(.*)",
        tornado.web.StaticFileHandler,
        {"path": os.path.dirname(__file__)},
    )

    handlers = [page_handler, socket_handler, static_handler, local_handler]

    settings = {
        "debug": True,
        "autoreload": False,
        "template_path": os.path.dirname(__file__) + "/templates",
    }
    EXCLUDE_LIST = (
        "width",
        "height",
    )

    def __init__(
        self,
        name,
        model_cls,
        visualization_elements,
        timesteps,
        model_params={},
        port=None,
        description="No description available",
    ):
        # Initializing the model
        self.model_name = name
        self.model_cls = model_cls
        if port:
            self.port = port
        self.description = description
        if hasattr(model_cls, "description"):
            self.description = model_cls.description
        elif model_cls.__doc__ is not None:
            self.description = model_cls.__doc__

        self.timesteps = timesteps
        self.model_kwargs = model_params

        """ Create a new visualization server with the given elements. """
        # Prep visualization elements:
        self.visualization_elements = visualization_elements
        self.package_includes = set()
        self.local_includes = set()
        self.js_code = []
        self.render_elements = []
        for element in self.visualization_elements:
            for include_file in element.package_includes:
                self.package_includes.add(include_file)
            for include_file in element.local_includes:
                self.local_includes.add(include_file)
            self.js_code.append(element.js_code)
            self.render_elements.append(element)

        # Initializing the application itself:
        super().__init__(self.handlers, **self.settings)

    @property
    def user_params(self):
        result = {}
        for param, val in self.model_kwargs.items():
            if isinstance(val, UserSettableParameter):
                result[param] = val.json

        return result

    def reset_model(self):
        """Reinstantiate the model object, using the current parameters."""

        model_params = {}
        for key, val in self.model_kwargs.items():
            if isinstance(val, UserSettableParameter):
                if (
                    val.param_type == "static_text"
                ):  # static_text is never used for setting params
                    continue
                model_params[key] = val.value
            else:
                model_params[key] = val

        self.model = self.model_cls(**model_params)

        for element in self.render_elements:
            element.reset()

    def set_background_variable(self, variable_name):
        self.model.artists.set_background_variable(variable_name)

    def render_model(self, update):
        """Turn the current state of the model into a dictionary of
        visualizations

        """
        return [element.render(self.model, update) for element in self.render_elements]

    def launch(self, port=None, browser=True):
        """Run the app."""
        if port is not None:
            self.port = port
        url = "http://127.0.0.1:{PORT}".format(PORT=self.port)
        print("Interface starting at {url}".format(url=url))
        self.listen(self.port, "0.0.0.0")

        if browser:
            webbrowser.open(url)
        tornado.ioloop.IOLoop.current().start()



# -*- coding: utf-8 -*-
import argparse
from honeybees import __name__

parser = argparse.ArgumentParser(
    description=f"{__name__.title()} is a package for high-speed geographical agent-based modelling in Python"
)
parser.add_argument(
    "--GUI",
    dest="GUI",
    action="store_true",
    help="The model can be run with or without a visual interface. The visual interface is useful to display the results in real-time while the model is running and to better understand what is going on. You can simply start or stop the model with the click of a buttion, or advance the model by an `x` number of timesteps. However, the visual interface is much slower than running the model without it.",
)
parser.set_defaults(GUI=False)
parser.add_argument(
    "--no-browser",
    dest="browser",
    action="store_false",
    help="Do not open browser when running the model. This option is, for example, useful when running the model on a server, and you would like to remotely access the model.",
)
parser.set_defaults(browser=True)
default_port = 8521
parser.add_argument(
    "--port",
    dest="port",
    type=int,
    default=8521,
    help=f"Port used for display environment (default: {default_port})",
)



# -*- coding: utf-8 -*-
"""This module is used to report data to the disk and to extract data for visualisation. After initialization of the reporter in the model, the :meth:`honeybees.report.Reporter.step` method is called every timestep, which can obtain data from the agents, and report to disk or save the data for visualiation or writing to disk once the model is finished. The variables to be reported can be configured in the config file. All data is saved to the report folder by default.

In the configuration file you can specify which data should be reported. In this file you can configure which data should be saved from the model in the `report` folder. This should be formatted as follows:

 - **report**:
    - **name**: name of the folder to which the data is saved.
        - **type**: agent type e.g., farmer. Should be identical to attribute name in Agents class.
        - **function**: whether to use a function to parse the data. 'null' means the data is saved literally, 'mean' takes the mean etc. Options are sum/mean.
        - **varname**: attribute name of variable in agent class.
        - **format**: format to save to. Can be 'csv' to save to csv-file per timestep, 'npy' to save in NumPy binary format, or 'npz' to save in NumPy compressed binary format.
        - **initial_only**: if true only save the data for the first timestep.
        - **save**: save variable in model run, or export, or both (save/save+export/export).
    - **name2**:
        - ...
        - ...
    - **...**:
        - ...
        - ...
"""

import re
from collections.abc import Iterable
import os
import numpy as np
import pandas as pd
from operator import attrgetter
from numba import njit
from math import isinf
from copy import deepcopy
import zarr
from numcodecs import Blosc
from typing import Union, Any

zstd_compressor = Blosc(cname="zstd", clevel=3, shuffle=Blosc.BITSHUFFLE)


def create_time_array(start, end, timestep, conf):
    if "frequency" not in conf:
        frequency = {"every": "day"}
    else:
        frequency = conf["frequency"]
    if "every" in frequency:
        every = frequency["every"]
        time = []
        current_time = start
        while current_time <= end:
            if every == "year":
                if (
                    frequency["month"] == current_time.month
                    and frequency["day"] == current_time.day
                ):
                    time.append(current_time)
            elif every == "month":
                if frequency["day"] == current_time.day:
                    time.append(current_time)
            elif every == "day":
                time.append(current_time)
            current_time += timestep
        return time
    elif frequency == "initial":
        return [start]
    elif frequency == "final":
        return [end]
    else:
        raise ValueError(f"Frequency {frequency} not recognized.")


class Reporter:
    """This class is used to report data to disk or for visualisation. The `step` method is called each timestep from the model.

    Args:
        model: The model.
        subfolder: Optional name of the subfolder to be reported in. By default the report folder from the configuration file is used (general:report_folder).
    """

    def __init__(self, model, folder: str) -> None:
        self.model = model
        if not hasattr(self.model, "agents"):  # ensure agents exist
            raise NameError(
                "Attribute agents of model does not exists. This most likely means that the reporter was created before the agents."
            )

        self.variables = {}
        self.timesteps = []
        self.export_folder = folder

        if (
            self.model.config is not None
            and "report" in self.model.config
            and self.model.config["report"] is not None
        ):
            self.enabled = True
            for name, conf in self.model.config["report"].items():
                if conf["format"] == "zarr":
                    filepath = os.path.join(self.export_folder, name + ".zarr.zip")
                    ds = zarr.open_group(filepath, mode="w")

                    time = create_time_array(
                        start=self.model.current_time,
                        end=self.model.end_time,
                        timestep=self.model.timestep_length,
                        conf=conf,
                    )

                    ds.create_dataset(
                        "time",
                        data=time,
                        dtype="datetime64[ns]",
                    )
                    ds["time"].attrs["_ARRAY_DIMENSIONS"] = ["time"]

                    conf["_file"] = ds
                    conf["_time_index"] = time
        else:
            self.enabled = False

        self.step()

    def check_value(self, value: Any):
        """Check whether the value is a Python integer or float, and is not infinite.

        Args:
            value: The value to be checked.
        """
        if not (
            isinstance(value, (int, float)) or value is None
        ):  # check item is normal Python float or int. This is required to succesfully convert to JSON.
            raise ValueError(
                f"value {value} of type {type(value)} is not Python float or int"
            )
        if isinstance(value, float):
            assert not isinf(value)

    def export_value(self, name: str, value: np.ndarray, conf: dict) -> None:
        """Exports an array of values to the export folder.

        Args:
            name: Name of the value to be exported.
            value: The array itself.
            conf: Configuration for saving the file. Contains options such a file format, and whether to export the array in this timestep at all.
        """
        if value is None:
            return None
        if "format" not in conf:
            raise ValueError(
                f"Export format must be specified for {name} in config file (npy/npz/csv/xlsx)."
            )
        if conf["format"] == "zarr":
            ds = conf["_file"]
            if name not in ds:
                if isinstance(value, (float, int)):
                    shape = (ds["time"].size,)
                    chunks = (1,)
                    compressor = None
                    dtype = type(value)
                    array_dimensions = ["time"]
                else:
                    shape = (ds["time"].size, value.size)
                    chunks = (1, value.size)
                    compressor = zstd_compressor
                    dtype = value.dtype
                    array_dimensions = ["time", "agents"]
                if dtype in (float, np.float32, np.float64):
                    fill_value = np.nan
                elif dtype in (int, np.int32, np.int64):
                    fill_value = -1
                else:
                    raise ValueError(
                        f"Value {dtype} of type {type(dtype)} not recognized."
                    )
                ds.create_dataset(
                    name,
                    shape=shape,
                    chunks=chunks,
                    dtype=dtype,
                    compressor=compressor,
                    fill_value=fill_value,
                )
                ds[name].attrs["_ARRAY_DIMENSIONS"] = array_dimensions
            index = conf["_time_index"].index(self.model.current_time)
            if value.size < ds[name][index].size:
                print("Padding array with NaNs or -1 - temporary solution")
                value = np.pad(
                    value,
                    (0, ds[name][index].size - value.size),
                    mode="constant",
                    constant_values=np.nan
                    if value.dtype in (float, np.float32, np.float64)
                    else -1,
                )
            ds[name][index] = value
        else:
            folder = os.path.join(self.export_folder, name)
            os.makedirs(folder, exist_ok=True)
            fn = f"{self.timesteps[-1].isoformat().replace('-', '').replace(':', '')}"
            if conf["format"] == "npy":
                fn += ".npy"
                fp = os.path.join(folder, fn)
                np.save(fp, value)
            elif conf["format"] == "npz":
                fn += ".npz"
                fp = os.path.join(folder, fn)
                np.savez_compressed(fp, data=value)
            elif conf["format"] == "csv":
                fn += ".csv"
                fp = os.path.join(folder, fn)
                if len(value) > 100_000:
                    self.model.logger.info(
                        f"Exporting {len(value)} items to csv. This might take a long time and take a lot of space. Consider using NumPy (compressed) binary format (npy/npz)."
                    )
                with open(fp, "w") as f:
                    f.write("\n".join([str(v) for v in value]))
            else:
                raise ValueError(f"{conf['format']} not recognized")

    def report_value(
        self, name: Union[str, tuple[str, Any]], value: Any, conf: dict
    ) -> None:
        """This method is used to save and/or export model values.

        Args:
            name: Name of the value to be exported.
            value: The array itself.
            conf: Configuration for saving the file. Contains options such a file format, and whether to export the data or save the data in the model.
        """
        if isinstance(value, list):
            value = [v.item() for v in value]
            for v in value:
                self.check_value(v)

        if "save" in conf:
            if conf["save"] not in ("save", "export"):
                raise ValueError(
                    f"Save type for {name} in config file must be 'save', 'save+export' or 'export')."
                )
            import warnings

            warnings.warn(
                "The `save` option is deprecated and will be removed in future versions. If you use 'save: export' the option can simply be removed (new default). If you use 'save: save', please replace with 'single_file: true'",
                DeprecationWarning,
            )
            if conf["save"] == "save":
                conf["single_file"] = True
            del conf["save"]

        if (
            "single_file" in conf
            and conf["single_file"] is True
            and conf["format"] != "zarr"  # for zarr, we always save per timestep
        ):
            try:
                if isinstance(name, tuple):
                    name, ID = name
                    if name not in self.variables:
                        self.variables[name] = {}
                    if ID not in self.variables[name]:
                        self.variables[name][ID] = []
                    self.variables[name][ID].append(value)
                else:
                    if name not in self.variables:
                        self.variables[name] = []
                    self.variables[name].append(value)
            except KeyError:
                raise KeyError(
                    f"Variable {name} not initialized. This likely means that an agent is reporting for a group that was not is not the reporter"
                )

        else:
            if "frequency" in conf and conf["frequency"] is not None:
                if conf["frequency"] == "initial":
                    if self.model.current_timestep == 0:
                        self.export_value(name, value, conf)
                elif conf["frequency"] == "final":
                    if self.model.current_timestep == self.model.n_timesteps:
                        self.export_value(name, value, conf)
                elif "every" in conf["frequency"]:
                    every = conf["frequency"]["every"]
                    if every == "year":
                        month = conf["frequency"]["month"]
                        day = conf["frequency"]["day"]
                        if (
                            self.model.current_time.month == month
                            and self.model.current_time.day == day
                        ):
                            self.export_value(name, value, conf)
                    elif every == "month":
                        day = conf["frequency"]["day"]
                        if self.model.current_time.day == day:
                            self.export_value(name, value, conf)
                    elif every == "day":
                        self.export_value(name, value, conf)
                    else:
                        raise ValueError(
                            f"Frequency every {conf['every']} not recognized (must be 'yearly', or 'monthly')."
                        )
                else:
                    raise ValueError(f"Frequency {conf['frequency']} not recognized.")
            else:
                self.export_value(name, value, conf)

    @staticmethod
    @njit(cache=True)
    def mean_per_ID(
        values: np.ndarray, group_ids: np.ndarray, n_groups: int
    ) -> np.ndarray:
        """Calculates the mean value per group.

        Args:
            values: Numpy array of values.
            group_ids: Group IDs for each value. Must be same size as values.
            n_groups: The total number of groups.

        Returns:
            mean_per_ID: The mean value for each of the groups.
        """
        assert values.size == group_ids.size
        size = values.size
        count_per_group = np.zeros(n_groups, dtype=np.int32)
        sum_per_group = np.zeros(n_groups, dtype=values.dtype)
        for i in range(size):
            group_id = group_ids[i]
            assert group_id < n_groups
            count_per_group[group_id] += 1
            sum_per_group[group_id] += values[i]
        return sum_per_group / count_per_group

    @staticmethod
    @njit(cache=True)
    def sum_per_ID(
        values: np.ndarray, group_ids: np.ndarray, n_groups: int
    ) -> np.ndarray:
        """Calculates the sum value per group.

        Args:
            values: Numpy array of values.
            group_ids: Group IDs for each value. Must be same size as values.
            n_groups: The total number of groups.

        Returns:
            sum_per_ID: The sum value for each of the groups.
        """
        assert values.size == group_ids.size
        size = values.size
        sum_per_group = np.zeros(n_groups, dtype=values.dtype)
        for i in range(size):
            group_id = group_ids[i]
            assert group_id < n_groups
            sum_per_group[group_id] += values[i]
        return sum_per_group

    def parse_agent_data(self, name: str, values: Any, agents, conf: dict) -> None:
        """This method is used to apply the relevant function to the given data.

        Args:
            name: Name of the data to report.
            values: Numpy array of values.
            agents: The relevant agent class.
            conf: Dictionary with report configuration for values.
        """
        function = conf["function"]
        if function is None or values is None:
            values = deepcopy(
                values
            )  # need to copy item, because values are passed without applying any a function.
            self.report_value(name, values, conf)
        else:
            function, *args = conf["function"].split(",")
            if function == "mean":
                value = np.mean(values)
                self.report_value(name, value, conf)
            elif function == "sum":
                value = np.sum(values)
                self.report_value(name, value, conf)
            elif function == "sample":
                sample = getattr(agents, "sample")
                value = values[sample]
                for s, v in zip(sample, value):
                    self.report_value((name, s), v, conf)
            elif function == "groupcount":
                for group in args:
                    group = int(group)
                    self.report_value(
                        (name, group), np.count_nonzero(values == group), conf
                    )
            else:
                raise ValueError(f"{function} function unknown")

    def extract_agent_data(self, name: str, conf: dict) -> None:
        """This method is used to extract agent data and apply the relevant function to the given data.

        Args:
            name: Name of the data to report.
            conf: Dictionary with report configuration for values.
        """
        agents = attrgetter(conf["type"])(self.model.agents)
        varname = conf["varname"]
        fancy_index = re.search(r"\[.*?\]", varname)
        if fancy_index:
            fancy_index = fancy_index.group(0)
            varname = varname.replace(fancy_index, "")

        try:
            values = getattr(agents, varname)
        except AttributeError:
            print(
                f"Trying to export '{varname}', but no such attribute exists for agent type '{conf['type']}'"
            )
            values = None

        if fancy_index:
            values = eval(f"values{fancy_index}")

        if "split" in conf and conf["split"]:
            for ID, admin_values in zip(agents.ids, values):
                self.parse_agent_data((name, ID), admin_values, agents, conf)
        else:
            self.parse_agent_data(name, values, agents, conf)

    def step(self) -> None:
        """This method is called every timestep. First appends the current model time to the list of times for the reporter. Then iterates through the data to be reported on and calls the extract_agent_data method for each of them."""
        self.timesteps.append(self.model.current_time)
        if self.enabled:
            for name, conf in self.model.config["report"].items():
                self.extract_agent_data(name, conf)

    def report(self) -> dict:
        """This method can be called to save the data that is currently saved in memory to disk."""
        report_dict = {}
        for name, values in self.variables.items():
            if isinstance(values, dict):
                df = pd.DataFrame.from_dict(values)
                df.index = self.timesteps
            elif isinstance(values[0], Iterable):
                df = pd.DataFrame.from_dict(
                    {k: v for k, v in zip(self.timesteps, values)}
                )
            else:
                df = pd.DataFrame(values, index=self.timesteps, columns=[name])
            if "format" not in self.model.config["report"][name]:
                raise ValueError(
                    f"Key 'format' not specified in config file for {name}"
                )
            export_format = self.model.config["report"][name]["format"]
            filepath = os.path.join(self.export_folder, name + "." + export_format)
            if export_format == "csv":
                df.to_csv(filepath)
            elif export_format == "xlsx":
                df.to_excel(filepath)
            elif export_format == "npy":
                np.save(filepath, df.values)
            elif export_format == "npz":
                np.savez_compressed(filepath, data=df.values)
            else:
                raise ValueError(f"save_to format {export_format} unknown")
        return report_dict



# -*- coding: utf-8 -*-
from random import random
from numba import njit


@njit
def bernoulli(p: float) -> bool:
    """Takes value from bernoulli distribution.

    Args:
        p: Probability

    Returns:
        outcome: True or False
    """
    assert 0 <= p <= 1
    return random() < p



# -*- coding: utf-8 -*-
import math
from honeybees.visualization.ModularVisualization import VisualizationElement
import numpy as np
from PIL import Image
import base64
from io import BytesIO


class Canvas(VisualizationElement):
    package_includes = ["canvas.js"]
    portrayal_method = None

    def __init__(self, max_canvas_width, max_canvas_height, unit="degrees"):
        """
        Instantiate a new Canvas
        """
        self.max_canvas_width = max_canvas_width
        self.max_canvas_height = max_canvas_height
        self.unit = unit
        new_element = "new Simple_Continuous_Module({}, {})".format(
            self.max_canvas_width, self.max_canvas_height
        )
        self.js_code = "elements.push(" + new_element + ");"

    def get_canvas_size(self, model):
        horizontal_distance = abs(model.xmin - model.xmax)
        vertical_distance = abs(model.ymin - model.ymax)

        if self.unit == "degrees":
            lat = (model.ymax + model.ymin) / 2
            horizontal_distance *= math.cos(math.radians(lat))
        elif self.unit == "meters":
            pass
        else:
            raise ValueError("unit must be degrees or meters")

        scaling_factor_width = self.max_canvas_width / (horizontal_distance)
        maybe_height = (vertical_distance) * scaling_factor_width
        if maybe_height > self.max_canvas_height:
            height = self.max_canvas_height
            scaling_factor_height = self.max_canvas_height / (vertical_distance)
            width = (horizontal_distance) * scaling_factor_height
        else:
            height = maybe_height
            width = self.max_canvas_width

        # +30 for legend
        if width + 30 < self.max_canvas_width:
            width += 30
        else:
            height /= self.max_canvas_width / width
            width = self.max_canvas_width

        return width, height

    def convert_lon_to_x(self, x, model):
        assert not np.isnan(x)
        return (x - model.area.geoms["xmin"]) / (
            model.area.geoms["xmax"] - model.area.geoms["xmin"]
        )

    def convert_lat_to_y(self, y, model):
        assert not np.isnan(y)
        return (y - model.area.geoms["ymin"]) / (
            model.area.geoms["ymax"] - model.area.geoms["ymin"]
        )

    def get_portrayal(self, artist, geography_kwargs):
        if geography_kwargs:
            portrayal = artist(**geography_kwargs)
        else:
            try:
                portrayal = artist()
            except TypeError as e:
                raise TypeError(
                    e.__repr__()
                    + "\nIs the missing argument specified in your config file?"
                )
        return portrayal

    def parse_linestring(self, coordinates, artist, geography_kwargs, model):
        portrayal = self.get_portrayal(artist, geography_kwargs)
        portrayal["lines"] = []
        lons = [coord[0] for coord in coordinates]
        lats = [coord[1] for coord in coordinates]
        portrayal["lines"].append(
            {
                "xs": [self.convert_lon_to_x(lon, model) for lon in lons],
                "ys": [self.convert_lat_to_y(lat, model) for lat in lats],
            }
        )
        return portrayal

    def parse_multilinestring(self, coordinates, artist, geography_kwargs, model):
        portrayal = self.get_portrayal(artist, geography_kwargs)
        portrayal["lines"] = []
        for line in coordinates:
            lons = [coord[0] for coord in line]
            lats = [coord[1] for coord in line]
            portrayal["lines"].append(
                {
                    "xs": [self.convert_lon_to_x(lon, model) for lon in lons],
                    "ys": [self.convert_lat_to_y(lat, model) for lat in lats],
                }
            )
        return portrayal

    def parse_multipolygon(self, coordinates, artist, geography_kwargs, model):
        portrayal = self.get_portrayal(artist, geography_kwargs)
        portrayal["rings"] = []
        for polygon in coordinates:
            for ring in polygon:
                lons = [coord[0] for coord in ring]
                lats = [coord[1] for coord in ring]
                portrayal["rings"].append(
                    {
                        "xs": [self.convert_lon_to_x(lon, model) for lon in lons],
                        "ys": [self.convert_lat_to_y(lat, model) for lat in lats],
                    }
                )
        return portrayal

    def parse_polygon(self, coordinates, artist, geography_kwargs, model):
        portrayal = self.get_portrayal(artist, geography_kwargs)
        portrayal["rings"] = []
        for ring in coordinates:
            lons = [coord[0] for coord in ring]
            lats = [coord[1] for coord in ring]
            portrayal["rings"].append(
                {
                    "xs": [self.convert_lon_to_x(lon, model) for lon in lons],
                    "ys": [self.convert_lat_to_y(lat, model) for lat in lats],
                }
            )
        return portrayal

    def parse_geojson(self, feature, *args, **kwargs):
        if "geometry" in feature:
            geometry = feature["geometry"]
        else:
            geometry = feature
        if geometry["type"] == "MultiPolygon":
            return self.parse_multipolygon(geometry["coordinates"], *args, **kwargs)
        elif geometry["type"] == "Polygon":
            return self.parse_polygon(geometry["coordinates"], *args, **kwargs)
        elif geometry["type"] == "LineString":
            return self.parse_linestring(geometry["coordinates"], *args, **kwargs)
        elif geometry["type"] == "MultiLineString":
            return self.parse_multilinestring(geometry["coordinates"], *args, **kwargs)
        else:
            raise NotImplementedError

    def get_static_space_state(self, model):
        if not hasattr(self, "static_space_state"):
            self.static_space_state = {}
            if (
                "draw_geography" in model.config["draw"]
                and model.config["draw"]["draw_geography"]
            ):
                for name, geography_kwargs in model.config["draw"][
                    "draw_geography"
                ].items():
                    used_IDs = set()
                    artist = getattr(model.artists, f"draw_{name}")
                    try:
                        geojsons = model.area.geoms[name]
                    except KeyError:
                        raise KeyError(
                            f"{name} geom not available in model.area.geoms. Did you pass it as a key to study_area?"
                        )

                    type_static_state_space = {}
                    if isinstance(geojsons, (list, tuple)):
                        for geojson in geojsons:
                            assert isinstance(geojson, dict)
                            if "properties" not in geojson:
                                raise ValueError(
                                    f"Geojson for {name} has not properties"
                                )
                            if "id" not in geojson["properties"]:
                                raise ValueError(f"Geojson for {name} has not ID")
                            ID = geojson["properties"]["id"]
                            if ID in used_IDs:
                                raise ValueError(
                                    f"ID is already used for drawing {name}. Each geojson must have a unique ID."
                                )
                            else:
                                used_IDs.add(ID)
                            type_static_state_space[ID] = self.parse_geojson(
                                geojson, artist, geography_kwargs, model
                            )

                    elif isinstance(geojsons, dict):
                        geojson = geojsons  # only 1 geojson available
                        if "properties" in geojson and "id" in geojson["properties"]:
                            ID = geojsons[ID]
                        else:
                            ID = None
                        type_static_state_space[ID] = self.parse_geojson(
                            geojson, artist, geography_kwargs, model
                        )
                    else:
                        raise ValueError

                    self.static_space_state[name] = type_static_state_space
            return self.static_space_state
        else:
            return self.static_space_state

    def get_agents_portrayals(
        self, model, artist, artist_kwargs, agent_subclass_id, agent_space_state, agents
    ):
        if artist_kwargs is None:
            artist_kwargs = {}
        locations = agents.locations.data
        if isinstance(locations, tuple):
            portrayal = artist(model, agents, **artist_kwargs)
            portrayal["x"] = self.convert_lon_to_x(locations[0], model)
            portrayal["y"] = self.convert_lat_to_y(locations[1], model)
            agent_space_state[f"{agent_subclass_id}"] = portrayal
        elif isinstance(locations, np.ndarray):
            artist_kwargs_to_pass = dict(artist_kwargs)
            if "draw_every_nth" in artist_kwargs_to_pass:
                draw_every_nth = artist_kwargs_to_pass.pop("draw_every_nth")
            else:
                draw_every_nth = 1
            agents_to_draw = np.arange(0, agents.n, draw_every_nth)
            if agents_to_draw.size > 100_000:
                print(
                    f"Drawing {agents_to_draw.size} agents. This might take a while. Consider setting `draw_every_nth_agent` to reduce the number of agents drawn. (This warning is displayed because more than 100.000 agents are to be drawn.)"
                )
            for idx in agents_to_draw:
                portrayal = artist(model, agents, idx, **artist_kwargs_to_pass)
                portrayal["x"] = self.convert_lon_to_x(locations[idx, 0], model)
                portrayal["y"] = self.convert_lat_to_y(locations[idx, 1], model)
                agent_space_state[f"{agent_subclass_id}_{idx}"] = (
                    portrayal  # converting to python int, as this is JSON serializable
                )
        else:
            raise ValueError(
                f"locations property of agent class {type(agents).__name__} must be tuple or (n, 2) Numpy array with locations"
            )

        return agent_space_state

    def render(self, model, update):
        agent_space_state = {}
        if model.config["draw"]["draw_agents"]:
            for agent_type, artist_kwargs in model.config["draw"][
                "draw_agents"
            ].items():
                artist_name = f'draw_{agent_type.replace(".", "_")}'
                try:
                    artist = getattr(model.artists, artist_name)
                except AttributeError:
                    raise AttributeError(
                        f"{agent_type} specified in initialization file has no corresponding method draw function {artist_name} in {model.artists}"
                    )
                if "." in agent_type:
                    agent_class, agent_subclass = agent_type.split(".")
                    agent_class = getattr(model.agents, agent_class)
                    agent_subclass = getattr(agent_class, agent_subclass)
                    for agent_subclass_id, agents in agent_subclass.items():
                        agent_space_state = self.get_agents_portrayals(
                            model=model,
                            artist=artist,
                            artist_kwargs=artist_kwargs,
                            agent_subclass_id=agent_subclass_id,
                            agent_space_state=agent_space_state,
                            agents=agents,
                        )
                else:
                    agents = getattr(model.agents, agent_type)
                    agent_space_state = self.get_agents_portrayals(
                        model=model,
                        artist=artist,
                        artist_kwargs=artist_kwargs,
                        agent_subclass_id="",
                        agent_space_state=agent_space_state,
                        agents=agents,
                    )

        space_state = self.get_static_space_state(model)
        space_state["agents"] = agent_space_state

        space_state_list = []
        background = model.artists.get_background()
        if background is not None:
            background, legend = background
            if isinstance(background, np.ndarray):
                img = Image.fromarray(background, mode="RGBA")
                buffer = BytesIO()
                img.save(buffer, format="PNG")

                space_state_list.append(
                    {
                        "type": "background",
                        "ysize": background.shape[0],
                        "xsize": background.shape[1],
                        "img": "data:image/png;base64,"
                        + base64.b64encode(buffer.getvalue()).decode("utf-8"),
                    }
                )
            else:
                raise NotImplementedError

            # legend for the background
            if legend["type"] == "colorbar":
                space_state_list.append(
                    {
                        "type": "colorbar",
                        "location": "right",
                        "min": legend["min"],
                        "max": legend["max"],
                        "unit": legend["unit"],
                        "color_min": model.artists.add_alpha_to_hex_color(
                            legend["color"], legend["min_colorbar_alpha"]
                        ),
                        "color_max": legend["color"] + "FF",  # hex 255
                    }
                )
            elif legend["type"] == "legend":
                assert "labels" in legend
                legend["location"] = "right"
                space_state_list.append(legend)
            else:
                raise NotImplementedError

        background_variables = model.artists.get_background_variables()
        if background_variables:
            space_state_list.append(
                {
                    "type": "background options",
                    "options": background_variables,
                    "currentselection": (
                        model.artists.background_variable
                        if hasattr(model.artists, "background_variable")
                        else None
                    ),
                }
            )

        # legend for the agents
        if hasattr(model.artists, "legend"):
            space_state_list.append(
                {
                    "type": "legend",
                    "location": "left",
                    "labels": model.artists.legend,
                }
            )

        for type_, type_portrayals in space_state.items():
            try:
                updater = getattr(model.artists, f"update_{type_}")
            except AttributeError:
                pass
            else:
                for ID, portrayal in type_portrayals.items():
                    updater(ID, portrayal)
            for portrayal in type_portrayals.values():
                assert isinstance(portrayal, dict)
                space_state_list.append(portrayal)

        return {"size": self.get_canvas_size(model), "draw": space_state_list}



# -*- coding: utf-8 -*-
from datetime import datetime
from typing import Any

RADIUS_EARTH_EQUATOR = 40075017  # m
distance_1_degree_latitude = RADIUS_EARTH_EQUATOR / 360  # m


def timeprint(*args, **kwargs: Any) -> None:
    """This function prints the current time in isoformat, followed by the normal print.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """
    print(datetime.now().isoformat(), *args, **kwargs)



# -*- coding: utf-8 -*-
"""Submodule that implements several classes for reading maps (image type and NetCDF type)."""

import datetime
import numpy as np
from math import ceil, floor
import xarray as xr
from honeybees.library.raster import sample_from_map
import rasterio
from rasterio import mask
from rasterio.windows import Window
from inspect import getcallargs
from typing import Union


class Reader:
    """This is a base class to read mapdata from files. The class takes as input the bounds of the relevant area. Any input maps are automatically cut to only read the required area. The class contains several functions to make reading data per agent easy and fast.

    Args:
        bounds: tuple of xmin, xmax, ymin, ymax of study_area.
    """

    def __init__(self, xmin: float, ymin: float, xmax: float, ymax: float) -> None:
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

    def get_source_gt(self) -> None:
        """Get Geotransformation of the source file. Must be implemented in the child class."""
        raise NotImplementedError

    def set_window_and_gt(self) -> None:
        """This function gets the geotransformation (gt) of the cut map, and the rows to read from the original data (rowslice, colslice)."""
        gt_ds = self.get_source_gt()

        colmin = floor((self.xmin - gt_ds[0]) / gt_ds[1])
        colmax = ceil((self.xmax - gt_ds[0]) / gt_ds[1])
        self.colslice = slice(colmin, colmax)

        rowmin = (self.ymax - gt_ds[3]) / gt_ds[5]
        rowmax = (self.ymin - gt_ds[3]) / gt_ds[5]
        # Detect if array is flipped upside down. If so, reverse rowmin and rowmax and use self.flipup to flip array when returned
        if rowmin > rowmax:
            rowmin, rowmax = rowmax, rowmin
            self.flipud = True
        else:
            self.flipud = False

        rowmin = floor(rowmin)
        rowmax = ceil(rowmax)

        self.rowslice = slice(rowmin, rowmax)

        assert colmin >= 0
        assert rowmin >= 0

        if self.flipud:
            self.gt = (
                gt_ds[0] + colmin * gt_ds[1],
                gt_ds[1],
                gt_ds[2],
                gt_ds[3] + rowmax * gt_ds[5],
                gt_ds[4],
                -gt_ds[5],
            )
        else:
            self.gt = (
                gt_ds[0] + colmin * gt_ds[1],
                gt_ds[1],
                gt_ds[2],
                gt_ds[3] + rowmin * gt_ds[5],
                gt_ds[4],
                gt_ds[5],
            )

    def sample_coords(self, coords: np.ndarray, *args, **kwargs) -> np.ndarray:
        """
        Args:
            coords: 2D NumPy array of coordinates to sample. First dimension is a list of coordinates, second dimension are the x and y.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            data: values as read from map for given coordinates.
        """
        array = self.get_data_array(*args, **kwargs)
        return sample_from_map(array, coords, self.gt)

    def get_data_array(self, *args, cache: bool = False, **kwargs):
        """Read data array from map.

        Parameters:
            cache (bool): cache output from previous call to function, check parameters (default: False)
            args: passed to child _get_data_array
            kwargs: passed to child _get_data_array

        Returns:
            array: requested array
        """
        if cache:
            call_args = getcallargs(self._get_data_array, *args, **kwargs)
            call_args.pop("self")
            if "kwargs" in call_args:
                call_args.update(call_args.pop("kwargs"))
            if (
                hasattr(self, "get_data_array_cache_args")
                and call_args == self.get_data_array_cache_args
            ):
                return self.get_data_array_cache

        array = self._get_data_array(*args, **kwargs)
        if cache:
            self.get_data_array_cache = array
            self.get_data_array_cache_args = call_args
        return array

    def delete_get_data_array_cache(self) -> None:
        """If results from get_data_array were cached, these can be deleted using this function."""
        if hasattr(self, "get_data_array_cache"):
            delattr(self, "get_data_array_cache")
            delattr(self, "get_data_array_cache_args")


class MapReader(Reader):
    """This class can be used to read image type data from files, and then to easily sample data from it.

    Args:
        fp: Filepath of file.
        bounds: tuple of xmin, xmax, ymin, ymax of study_area.
    """

    def __init__(
        self, fp: str, xmin: float, ymin: float, xmax: float, ymax: float
    ) -> None:
        Reader.__init__(self, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
        self.fp = fp
        self.ds = rasterio.open(fp, "r")
        self.set_window_and_gt()

    def get_source_gt(self) -> tuple[float, float, float, float, float, float]:
        """Get geotransformation of the source file.

        Returns:
            gt: Geotransformation of source file.
        """
        return self.ds.transform.to_gdal()

    @property
    def source_shape(self) -> tuple[int, int]:
        """Get shape of source file.

        Returns:
            height: height of array.
            width: width of array.
        """
        return self.ds.shape

    def _get_data_array(self) -> np.ndarray:
        """Read array from file for study_area.

        Returns:
            data: Array of data for study_area.
        """
        data = self.ds.read(window=Window.from_slices(self.rowslice, self.colslice))
        if data.shape[0] == 1:
            return data[0]
        else:
            return data

    def sample_geom(
        self, geom: dict, all_touched=False, nodata: Union[float, int] = -1
    ) -> np.ndarray:
        """Sample geometry from array.

        Args:
            geom: GeoJSON dictionary of geometry.
            nodata: Value to fill nodata.
            all_touched: Bool to include a pixel in the mask if it touches any of the shapes.
        Returns:
            data: Values for geometry.
        """
        geom = geom["geometry"] if "geometry" in geom else geom
        data, transform = mask.mask(
            self.ds, [geom], crop=True, all_touched=all_touched, nodata=nodata
        )
        return data[0, :, :]

    def sample_geoms(self, geoms: list[dict], nodata: int = -1):
        """Sample multiple geometries from data array.

        Args:
            geoms: List of GeoJSON dictionary of geometries.
            nodata: Value to fill nodata.

        Yields:
            data: Values for geometry.
        """
        for geom in geoms:
            yield self.sample_geom(geom=geom, nodata=nodata)


class NetCDFReader(Reader):
    def __init__(
        self,
        fp: str,
        varname: str,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
        latname: str = "lat",
        lonname: str = "lon",
        timename: str = "time",
    ) -> None:
        """This class can be used to read data from NetCDF files, and then to easily sample data from it.

        Args:
            fp: Filepath of file.
            varname: netcdf variable to read
            bounds: tuple of xmin, xmax, ymin, ymax of study_area.
            latname: name of latitude variable
            lonname: name of longitude variable
            timename: name of the time variable
        """
        self.fp = fp
        self.latname = latname
        self.lonname = lonname
        self.timename = timename

        Reader.__init__(self, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
        self.ds = xr.open_dataset(self.fp)

        try:
            self.data = self.ds.variables[varname]
        except KeyError:
            raise KeyError(
                f'{varname} not found in NetCDF. Possible variables are {", ".join(self.ds.variables.keys())}'
            )

        self.set_window_and_gt()
        self.get_current_time_index()

    @property
    def title(self) -> str:
        """Title of NetCDF file."""
        return self.ds.title

    @property
    def source_shape(self):
        """Get shape of source file.

        Returns:
            height: height of array.
            width: width of array.
        """
        return (
            self.ds.variables[self.latname].size,
            self.ds.variables[self.lonname].size,
        )

    def get_source_gt(self):
        """Construct a GDAL-style Geotransformation from the source NetCDF using the longitude and latitude data.

        Returns:
            gt: Geotransformation of source file.
        """
        gt = (
            self.ds.rename({self.lonname: "x", self.latname: "y"})
            .rio.transform()
            .to_gdal()
        )
        assert gt[2] == 0 and gt[4] == 0, "Geotransformation must have no rotation."
        return gt

    def get_current_time_index(self) -> None:
        """Read the time indices from NetCDF file in Python datetime format. If time index does not exist, it is not set."""
        if self.timename in self.ds.variables:
            self.time_index = self.ds[self.timename].to_index()
        else:
            self.time_index = None

    def _get_data_array(self, dt: datetime.datetime = None) -> np.ndarray:
        """
        Read data array from NetCDF file for study_area.

        Args:
            dt: Datetime to read. Required if data has time dimension, otherwise should not be given.

        Returns:
            data: Array of data for study_area.
        """
        if self.time_index is not None:
            if not dt:
                raise ValueError(
                    "Must specify datetime for NetCDF with time component by passing dt-argument to the function call."
                )
            time_index = self.time_index.get_loc(dt)
            array = self.data.isel(
                {
                    self.timename: time_index,
                    self.latname: self.rowslice,
                    self.lonname: self.colslice,
                }
            )
        else:
            assert dt is None
            array = self.data.isel(
                {self.latname: self.rowslice, self.lonname: self.colslice}
            )
            # array = self.data[self.rowslice, self.colslice]

        assert array.size > 0
        if self.flipud:
            return np.flipud(array)
        else:
            return array.data

    def close(self) -> None:
        """Close NetCDF file."""
        self.ds.close()


class ArrayReader(Reader):
    def __init__(
        self,
        array: str,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
        transform: Union[
            rasterio.Affine, tuple[float, float, float, float, float, float]
        ],
    ) -> None:
        Reader.__init__(self, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
        self.array = array
        if isinstance(transform, rasterio.Affine):
            self.transform = transform.to_gdal()
        else:
            self.transform = transform
        self.set_window_and_gt()

    def get_source_gt(self) -> tuple[float, float, float, float, float, float]:
        """Get geotransformation of the source file.

        Returns:
            gt: Geotransformation of source file.
        """
        return self.transform

    @property
    def source_shape(self) -> tuple[int, int]:
        """Get shape of source file.

        Returns:
            height: height of array.
            width: width of array.
        """
        return self.array.shape

    def _get_data_array(self) -> np.ndarray:
        """Read array from file for study_area.

        Returns:
            data: Array of data for study_area.
        """
        data = self.array[self.rowslice, self.colslice]
        if data.shape[0] == 1:
            return data[0]
        else:
            return data

    def sample_geom(
        self, geom: dict, all_touched=False, nodata: Union[float, int] = -1
    ) -> np.ndarray:
        raise NotImplementedError

    def sample_geoms(self, geoms: list[dict], nodata: int = -1):
        raise NotImplementedError



# -*- coding: utf-8 -*-
import colorsys
import math
import numpy as np


class Artists:
    """Test"""

    def __init__(self, model):
        self.model = model

    def draw_global(self):
        return {
            "type": "shape",
            "shape": "polygon",
            "filled": False,
            "color": "Black",
            "edge": True,
        }

    def select_latest_reporter_values(self, report_group, name):
        return self.model.reporter.variables["country"][name][-1]

    def add_alpha_to_hex_color(self, hex_color, alpha):
        decimal_alpha = int(alpha * 255)
        alpha_hex = hex(decimal_alpha).replace(
            "0x", ""
        )  # all hex values are prefixed with '0x'
        if len(alpha_hex) == 1:  # value has 1 number
            alpha_hex = "0" + alpha_hex
        return hex_color + alpha_hex

    def get_alpha(self, value, min, max):
        alpha = (value - min) / (max - min)
        if alpha < 0:
            alpha = 0
        if alpha > 1:
            alpha = 1
        return alpha

    def rgb_to_hex(self, rgb):
        def clamp(x):
            x = int(x * 255)
            return max(0, min(x, 255))

        hex_color = f"#{clamp(rgb[0]):02x}{clamp(rgb[1]):02x}{clamp(rgb[2]):02x}"
        if len(rgb) == 4:
            hex_color += f"{clamp(rgb[3]):02x}"
        return hex_color

    def hex_to_rgb(self, h):
        if h.startswith("#"):
            h = h.lstrip("#")
        return tuple(int(h[i : i + 2], 16) / 255 for i in (0, 2, 4))

    def generate_distinct_colors(self, n, mode):
        hsv_tuples = [(x * 1.0 / n, 0.5, 0.5) for x in range(n)]
        if mode == "hsv":
            return hsv_tuples
        rgb_tuples = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
        if mode == "rgb":
            return rgb_tuples
        if mode == "hex":
            return [self.rgb_to_hex(rgb) for rgb in rgb_tuples]
        else:
            raise NotImplementedError

    def generate_discrete_colors(self, n, color, mode, min_alpha=0):
        alphas = np.linspace(min_alpha, 1, n)
        if mode == "hex":
            return [self.add_alpha_to_hex_color(color, alpha) for alpha in alphas]
        elif mode == "rgb":
            return [(color[0], color[1], color[2], alpha) for alpha in alphas]
        else:
            raise NotImplementedError

    def round_to_n_significant_digits(self, value, n):
        if value == 0:
            return 0
        else:
            if math.isinf(value) or math.isnan(value):
                return value
            else:
                return round(value, n - int(math.floor(math.log10(abs(value)))) - 1)

    def draw_river(self):
        return {"type": "shape", "shape": "line", "color": "Blue"}

    def get_background(self):
        return None

    def get_background_variables(self):
        return None


if __name__ == "__main__":
    artist = Artists(None)
    print(artist.generate_distinct_colors(10, mode="hex"))



# -*- coding: utf-8 -*-
import numpy as np


class AgentBaseClass:
    def __init__(self):
        pass

    @property
    def activation_order_random(self):
        """Returns containing all agent ids in random order.

        Returns:
            np.array: agent ids in random order
        """
        activation_order = np.arange(0, self.n, 1, dtype=np.int32)
        np.random.shuffle(activation_order)
        return activation_order

    def initiate_agents(self):
        raise NotImplementedError

    def step(self):
        raise NotImplementedError(f"step function not implemented in {self}")



# -*- coding: utf-8 -*-
from datetime import timedelta
from dateutil.relativedelta import relativedelta
import logging
import yaml


class Model:
    def __init__(
        self,
        current_time,
        timestep_length,
        config_path=None,
        n_timesteps=None,
        args=None,
    ):
        self._current_time = current_time
        self.timestep_length = timestep_length
        self.n_timesteps = n_timesteps
        if config_path is not None:
            self.config = self.setup_config(config_path)
        self.args = args
        self.logger = self.create_logger()
        self.logger.info("Initializing model")
        self.current_timestep = 0
        self.running = True

    @property
    def bounds(self):
        """Property returning the bounds of the current model as tuple

        Returns:
            tuple: xmin, xmax, ymin, ymax bounds of the model

        """
        return self.xmin, self.xmax, self.ymin, self.ymax

    @property
    def xmin(self):
        """Get model xmin

        Returns:
            float: xmin
        """
        return self.area.geoms["xmin"]

    @property
    def xmax(self):
        """Get model xmax

        Returns:
            float: xmax
        """
        return self.area.geoms["xmax"]

    @property
    def ymin(self):
        """Get model ymin

        Returns:
            float: ymin
        """
        return self.area.geoms["ymin"]

    @property
    def ymax(self):
        """Get model ymax

        Returns:
            float: ymax
        """
        return self.area.geoms["ymax"]

    @property
    def current_time_fmt(self):
        if self.timestep_length.days >= 1:
            dateformat = "%d %b %Y"
        else:
            dateformat = "%d %b %Y %H:%M"
        formatted_date = self.current_time.strftime(dateformat)
        if formatted_date.startswith(
            "0"
        ):  # Windows cannot handle %-d, so this is a safe way to remove the preceding 0
            formatted_date = formatted_date[1:]
        return formatted_date

    @property
    def current_time(self):
        """
        Returns:
            datetime.datetime: current model time
        """
        return self._current_time

    @property
    def end_time(self):
        """
        Returns:
            datetime.datetime: end time of the model
        """
        return self._current_time + self.n_timesteps * self.timestep_length

    @current_time.setter
    def current_time(self, v):
        self._current_time = v

    @property
    def current_timestep(self):
        """
        Returns:
            int: current model timestep
        """
        return self._current_timestep

    @current_timestep.setter
    def current_timestep(self, v):
        self._current_timestep = v

    def create_logger(self):
        logger = logging.getLogger("honeybees")

        if (
            self.config
            and "logging" in self.config
            and "loglevel" in self.config["logging"]
        ):
            loglevel = self.config["logging"]["loglevel"]
        else:
            loglevel = "INFO"
        logger.setLevel(logging.getLevelName(loglevel))

        if (
            self.config
            and "logging" in self.config
            and "logfile" in self.config["logging"]
        ):
            logfile = self.config["logging"]["logfile"]
        else:
            logfile = "honeybees.log"
        file_handler = logging.FileHandler(logfile, mode="w")
        logger.addHandler(file_handler)

        formatter = logging.Formatter("%(asctime)s : %(levelname)s : %(message)s")
        file_handler.setFormatter(formatter)

        return logger

    def setup_config(self, config):
        if config is None:
            return None
        elif isinstance(config, dict):
            return config
        elif isinstance(config, str):
            with open(config, "r") as f:
                config = yaml.load(f, Loader=yaml.FullLoader)
            return config
        else:
            raise ValueError(
                f"config should be a dict or a path to a yaml file, not {type(config)}"
            )

    def parse_step_str(self, step_string):
        if step_string == "day":
            new_time = self.current_time + timedelta(days=1)
            difference = new_time - self.current_time
            n = int(difference / self.timestep_length)
        elif step_string == "week":
            new_time = self.current_time + timedelta(days=7)
            difference = new_time - self.current_time
            n = int(difference / self.timestep_length)
        elif step_string == "month":
            new_time = self.current_time + relativedelta(months=1)
            difference = new_time - self.current_time
            n = int(difference / self.timestep_length)
        elif step_string == "year":
            new_time = self.current_time + relativedelta(years=1)
            if isinstance(self.timestep_length, relativedelta):
                assert self.timestep_length.years == 1
                n = 1
            else:
                difference = new_time - self.current_time
                n = int(difference / self.timestep_length)
        elif step_string == "decade":
            new_time = self.current_time + relativedelta(years=10)
            if isinstance(self.timestep_length, relativedelta):
                assert self.timestep_length.years == 1
                n = 10
            else:
                difference = new_time - self.current_time
                n = int(difference / self.timestep_length)
        elif step_string == "century":
            new_time = self.current_time + relativedelta(years=100)
            if isinstance(self.timestep_length, relativedelta):
                assert self.timestep_length.years == 1
                n = 100
            else:
                difference = new_time - self.current_time
                n = int(difference / self.timestep_length)
        else:
            raise ValueError(f"{step_string} not a known step_size")
        return n

    def step(self, step_size=1, report=True):
        if isinstance(step_size, str):
            n = self.parse_step_str(step_size)
        else:
            n = step_size
        self.current_time += self.timestep_length
        self.current_timestep += 1

        assert isinstance(n, int) and n > 0
        for _ in range(n):
            # t0 = time()
            # print('Simulating agent behavior')
            self.agents.step()
            # t1 = time()
            if report:
                self.reporter.step()
            # t2 = time()
            # print('\tstep time', t1- t0)
            # print('\treport time', t2 - t1)

    def run(self, report=True):
        for _ in range(self.n_timesteps):
            self.step(report=report)
        self.report()

    def report(self):
        return self.reporter.report()
